#!/usr/bin/python3

import argparse

from pwn import *
from bitstring import ConstBitStream

import hexdump


def u24(data):
	return u32(data + b"\x00")

class Node:
	def __init__(self):
		self.right = None
		self.left = None
		self.symbol = None

	def __str__(self):
		return "[%02.2x]" % (self.symbol)
		
def insert(node,symbol,nbits):
	if node.symbol is None:
		if nbits == 0:
			node.symbol = symbol
			return True
		else:
			searchNode = None
			if node.left == None:
				searchNode = Node()
				node.left = searchNode
			else:
				searchNode = node.left
				
			if not insert(searchNode,symbol,nbits - 1):
				searchNode = None
				if node.right == None:
					searchNode = Node()
					node.right = searchNode
				else:
					searchNode = node.right
					
				return insert(searchNode,symbol,nbits - 1)
			else:
				return True
				
		return False
	
def walk_tree(node,stream,bit_read=0):
	if node.symbol is not None:
		return (node.symbol,bit_read)
	else:
		bit = stream.read(1).uint
		bit_read += 1
		
		if bit == 0:
			if node.left is not None:
				return walk_tree(node.left,stream,bit_read)
			else:
				raise Exception("huffman exception")
		else:
			if node.right is not None:
				return walk_tree(node.right,stream,bit_read)
			else:
				raise Exception("huffman exception")

def huffman_decode_bytes(huffman_tree,data,size):
	bytes_list,bytes_read = huffman_decode_list(huffman_tree,data,size)
	return (bytes(bytes_list),bytes_read)
	
def huffman_decode_list(huffman_tree,data,size):
	result = []
	bitstream = ConstBitStream(data)
	bits_read = 0
	for i in range(0,size):
		c,bits = walk_tree(huffman_tree,bitstream)
		result.append(c)
		bits_read+=bits
	
	bytes_read = int(bits_read / 8) + int( (bits_read % 8) != 0 )
	return (result,bytes_read)
	
	
def huffman_decode_special(huffman_tree,data,size):
	result = []
	bitstream = ConstBitStream(data)
	bits_read = 0
	for i in range(0,size):
		c,bits = walk_tree(huffman_tree,bitstream)
		
		extra_bits = 0
		if c > 15:
			extra_bits = c - 0xC
		
		if extra_bits > 0:
			# print("read %d extra bits for symbol %d" % (extra_bits,c))
			c = bitstream.read(extra_bits).uint + (1 << extra_bits)  
		
		result.append(c)
		bits_read += bits + extra_bits
	
	bytes_read = int(bits_read / 8) + int( (bits_read % 8) != 0 )
	return (result,bytes_read)
	
def huffman_decode_special2(huffman_tree,data,size):
	result = []
	bitstream = ConstBitStream(data)
	bits_read = 0
	for i in range(0,size):
		c,bits = walk_tree(huffman_tree,bitstream)
		
		extra_bits = 0
		if c > 1:
			extra_bits = c - 1
		
		if extra_bits > 0:
			c = bitstream.read(extra_bits).uint + (1 << extra_bits)  
		
		result.append(c)
		bits_read += bits + extra_bits
	
	bytes_read = int(bits_read / 8) + int( (bits_read % 8) != 0 )
	return (result,bytes_read)
	

	
def build_huffman_tree(data,encoded_size=8):
	root = Node()
	
	bits_read = 0
	
	b = ConstBitStream(data)
	nsymbols = b.read(encoded_size).uint
	# print("nsymbols %d" % (nsymbols + 1))
	bits_read+= encoded_size
	
	for i in range(0,nsymbols+1):
		symbol = b.read(encoded_size).uint
		bitsize = b.read(4).uint + 1
		# print("symbol: %02.2x, bits : %d" % (symbol,bitsize))
		insert(root,symbol,bitsize)
		bits_read += encoded_size + 4
	
	bytes_read = int(bits_read / 8) + int( (bits_read % 8) != 0 )
	return (root,bytes_read)
	
	
ENC_SIZE = 5
def decompress_flux(data):
	
	uncompress_size = u24(data[0:3])
	encoded_size = u24(data[3:6])
	offset = 6
	
	print("uncompress_size : %d" % uncompress_size)
	print("encoded_size : %d" % encoded_size)
	
	tree,seek = build_huffman_tree(data[offset:])
	offset+= seek
	chain,seek = huffman_decode_bytes(tree,data[offset:],uncompress_size)
	offset+= seek
	
	array_length = u24(data[offset:offset + 3])
	print("size : %08.8x" % array_length)
	offset+= 3
	
	
	print("chain : %s" % chain[0:80])
	
	tree,seek = build_huffman_tree(data[offset:],encoded_size=ENC_SIZE)
	offset+=seek
	sym_offset,seek = huffman_decode_special(tree,data[offset:],array_length)
	offset+=seek
	print(sym_offset)
	
	tree,seek = build_huffman_tree(data[offset:],encoded_size=ENC_SIZE)
	offset+=seek
	sym_size,seek = huffman_decode_special2(tree,data[offset:],array_length)
	offset+=seek
	print(sym_size)
	
	tree,seek = build_huffman_tree(data[offset:],encoded_size=ENC_SIZE)
	offset+=seek
	repeat,seek = huffman_decode_special(tree,data[offset:],array_length)
	offset+=seek
	print(repeat)
	
	uncompressed = b""
	curr = 0
	
	for i in range(0,array_length):
		repeated = chain[curr + sym_offset[i] - sym_size[i]:curr + sym_offset[i]]
		if sym_size[i] != 0:
			n = int(repeat[i] / sym_size[i])
			uncompressed += chain[curr:curr + sym_offset[i]]
			uncompressed += repeated * n
			if (repeat[i] % sym_size[i])!=0:
				uncompressed += repeated[0:repeat[i]%sym_size[i]]
		else:
			uncompressed += chain[curr:curr + sym_offset[i]]
			
		curr+= sym_offset[i]
		
	print("chain length : %d" % len(chain))
	print("%d bytes decompressed" % len(uncompressed))
	return uncompressed[0:0x10000]



def decompress(fin):
	result = b""
	header = fin.read(3)
	while len(header) == 3:
		length = u24(header)
		flux = fin.read(length)
		result += decompress_flux(flux)	
		header = fin.read(3)
	return result

if __name__=='__main__':
	parser = argparse.ArgumentParser(description='decompress zz file')
	parser.add_argument('input',help='compressed file',type=str)
	parser.add_argument('--output',help='output file',type=str)
	args = parser.parse_args()
	
	with open(args.input,"rb") as fin:
		data = decompress(fin)
		if args.output:
			with open(args.output,"wb") as fout:
				fout.write(data)
		else:
			print(data)
			hexdump.hexdump(data)