import logging
from pwn import *
import coloredlogs

logger = logging.getLogger('FTP')
logger.setLevel(logging.DEBUG)
coloredlogs.install(level='DEBUG', logger=logger,fmt='%(name)s:%(levelname)s:%(message)s')

class FTP:
	def __init__(self,ip,port=31337):
		self.ip = ip
		self.p = remote(ip,port)
		self.recv_answer()
		
	def recv_answer(self,until=None):
		data = None
		if until is not None:
			data = self.p.recvuntil(until)
			logger.debug(data.decode('ascii'))
		else:
			data = self.p.recvline()
			logger.debug(data.decode('ascii').replace("\n",""))
		
		return data
		
	def cmd_user(self,username):
		self.p.sendline(b"USER %s" % username)
		self.recv_answer()
	
	def cmd_pass(self):
		self.p.sendline(b"PASS")
		self.recv_answer()
		
	def cmd_feat(self):
		self.p.sendline(b"FEAT")
		self.recv_answer(b"211 End\n")
		
	def cmd_dbg(self):
		self.p.sendline(b"DBG")
		self.recv_answer()
		
	def cmd_pasv(self):
		self.p.sendline(b"PASV")
		ip_str = self.recv_answer().split(b"(")[1].replace(b")",b"")
		a,b,c,d,port_high,port_low = [ int(e) for e in ip_str.split(b",")]
		
		if self.ip != "127.0.0.1":
			conn = remote("%d.%d.%d.%d" % (a,b,c,d),(port_high << 8)|port_low)
		else:
			conn = remote("127.0.0.1",33344)
		return conn
		
	def cmd_getpasv(self):
		self.p.sendline(b"PASV")
		ip_str = self.recv_answer().split(b"(")[1].replace(b")",b"")
		a,b,c,d,port_high,port_low = [ int(e) for e in ip_str.split(b",")]
		return port_high,port_low

	def cmd_list(self):
		conn = self.cmd_pasv()
		self.p.sendline(b"LIST")
		self.recv_answer()
		logger.info(conn.recvall().decode('ascii'))
		self.recv_answer()
		conn.close()
		
	def cmd_retr(self,filename,skip=False):
		conn = self.cmd_pasv()
		self.p.sendline(b"RETR %s" % filename)
		if skip:
			data = conn.recvall()
			conn.close()
			self.recv_answer()
			return data

		if b"Ok" in self.recv_answer():
			data = conn.recvall()
			self.recv_answer()
			conn.close()
		else:
			data = None
		return data
		
	def cmd_cert(self,certificate,exploit=False):
		sz = base64.b64encode(certificate)
		self.p.sendline(b"CERT %s" % sz)
		if not exploit:
			answer = self.recv_answer()
			return b"Ok" in answer
		else:
			return True

	def close(self):
		self.p.sendline(b"QUIT")
		self.recv_answer()
		self.p.close()
		